divert(-1)
#********************************************************************
#
#M4 definitions for implementing the Retargetable Concurrent Small C
#virtual machine on the Ubicom SX28
#Registers reflect 8086 registers.
#
# compile c program: rcsc program.c
# this will create program.m4
# compile m4 to sasm source: m4 sx28_16.m4 program.m4 > program.src
#
#Jan 6, 2005 Peter Verkaik
#
#********************************************************************
#
#RCSC pseudo instructions
#
#********************************************************************
define(`CSCEXTRN',`	;Retargetable Concurrent Small C compiler')

define(`CSCTERM',
`;extrn	_cscinit
	;MNTRSEG SEGMENT PUBLIC
	;dw	1a1ah
	;MNTRSEG ENDS
	;TASKSEG SEGMENT PUBLIC
	;dw	1a1ah
	;TASKSEG ENDS
	;INTRSEG SEGMENT PUBLIC
	;dw	1a1ah
	;INTRSEG ENDS')

define(`EXTERNAL',
`extrn $1:$2')

define(`TOSEG',
`define(`CURSEG', $1) ifelse($1, 1, 
`	;DATA SEGMENT PUBLIC',
`	;CODE SEGMENT PUBLIC')')

define(`ENDSEG',
`ifelse($1, 1,
`	;DATA ENDS',
`	;CODE ENDS')')

define(`DECLPUBLIC',
`ifelse(CURSEG, 2, 
`	;PUBLIC $1
$1', 
`	;PUBLIC $1
DECLARE($1)')')

define(`DECLARE', 
`define(`LABELON', 1)
DECLAREMEM($1,')dnl

define(`REFm', 
`define(`LABELON', 1)
REFmMEM($1,')dnl

define(`SHADOW', `')
define(`MODULENAME', `')
define(`BEGINLIT', `')
define(`ENDLIT', `')
define(`BEGINDUMP', `')dnl

#********************************************************************
#
#macros used internally
#
#********************************************************************
define(`repeat', 
`ifelse(eval($2 <= 0), 1, , 
`repeat(`$1', decr($2))'
$1)')

define(`DECLAREMEM', 
`define(`LABELON', 0)
ifelse($2, _WORD_, `_W_$1
$1	equ	ramptr
	dfw	', 
$2, _WORDn, `_W_$1
$1	equ	ramptr
	dfw	$3',
$2, _WORDr0, `_W_$1
$1	equ	ramptr
	dfw	$3',
$2, _BYTE_, `$1	equ	ramptr
	dfs	', 
$2, _BYTEn, `$1	equ	rampter
	dfs	$3',
`$1	equ	ramptr
	dfs	$3')')

define(`REFmMEM', 
`define(`LABELON', 0)
ifelse($2, _WORD_, `$1	dw	', 
$2, _WORDn, `$1	dw	$3',
$2, _WORDr0, `$1	dw	$3',
$2, _BYTE_, `$1	dw	', 
$2, _BYTEn, `$1	dw	$3',
`$1	dw	$3')')

#********************************************************************
#
#interrupt, task and monitor table creation macros
#
#********************************************************************
define(`INTRS', 
`	;INTRSEG SEGMENT PUBLIC
	dw	$1,$2
	;INTRSEG ENDS')

define(`TASKS', 
`	;TASKSEG SEGMENT PUBLIC
	dw	$1,$2,$3
	;TASKSEG ENDS')

define(`MNTRS', 
`	;MNTRSEG SEGMENT PUBLIC
	dw	$1
	;MNTRSEG ENDS')

#********************************************************************
#
#RCSC virtual machine instruction set
#
#********************************************************************
#
# This set replaces the 8086 16bit registers.
# Used registers with this set: AL,AH,BL,BH,CL,CH,DL,DH,BP,SP
# This code assumes bank CPU is accessable
# so any asm code must exit with bank CPU if ram bank changed.
# AX (AH:AL) and BX (BH:BL) are the main registers
# CX (CH:CL) is primarily used for addresses and pointers
# DX (DH:DL) is primarily used for values and pointer offsets 
# The code is optimized for space, meaning the least code words
# possible before calling a runtime routine.

define(`ADD12',
``;;ADD12'
	call	@add12')

define(`ADD1n',
``;;ADD1n($1)'
	mov	DL,#$1&255
	mov	DH,#$1>>8
	call	@add1n')

define(`ADD21',
``;;ADD21'
	call	@add21')

define(`ADD2n',
``;;ADD2n($1)'
	mov	DL,#$1&255
	mov	DH,#$1>>8
	call	@add2n')

define(`ADDbpn',
``;;ADDbpn($1)'
	mov	DL,#$1&255
	call	@addbpn')

define(`ADDwpn',
``;;ADDwpn($1)'
	mov	DL,#$1&255
	mov	DH,#$1>>8
	call	@addwpn')

define(`ADDm_',
`ADDm_COMMAn($1')

define(`ADDm_COMMAn',
``;;ADDm_COMMAn($1, $2)'
	mov	CL,#$1&255
	mov	DL,#$2&255
`ifdef' _W_`$1'
	mov	DH,#$2>>8
	call	@addwmn
`else'
	call	@addbmn
`endif'
')

define(`ADDSP',
``;;ADDSP($1)'
	add	SP,#$1&255')

define(`AND12',
``;;AND12'
	call	@and12')

define(`ANEG1',
``;;ANEG1'
	call	@aneg1')

define(`ARGCNTn',
``;;ARGCNTn($1)'
ifelse($1, 0, `	clr	CL', `	mov	CL,#$1&255')')

define(`ASL12',
``;;ASL12'
	call	@asl12')

define(`ASR12',
``;;ASR12'
	call	@asr12')

#note that we jump to callf but it acts as a call
#this way the return address in not placed on the call stack
#remember CL holds argument count
define(`CALL1',
``;;CALL1'
	mov	DL,#($+6)&255	;ret addr
	mov	DH,#($+4)>>8
	jmp	@callf')

#note that we jump to callf but it acts as a call
#this way the return address in not placed on the call stack
#remember CL holds argument count
define(`CALLm',
``;;CALLm($1)'
	mov	AL,#$1&255	;addr to call
	mov	AH,#$1>>8
	mov	DL,#($+6)&255	;ret addr
	mov	DH,#($+4)>>8
	jmp	@callf')

define(`BYTE_', 
`ifelse(LABELON, 1, `_BYTE_, $1)', `	dw	')')

define(`BYTEn', 
`ifelse(LABELON, 1, `_BYTEn, $1)', `	dw	$1')')

define(`BYTEr0', 
`ifelse(LABELON, 1, `_BYTEr0, $1)', `	dw	$1,0')')

define(`COM1',
``;;COM1'
	not	AL
	not	AH')

define(`COMMAn',
`,$1)')

define(`DBL1',
``;;DBL1'
	call	@dbl1')

define(`DBL2',
``;;DBL2'
	call	@dbl2')

define(`DECbp',
``;;DECbp'
	call	@decbp')

define(`DECwp',
``;;DECwp'
	call	@decbp')

define(`DIV12',
``;;DIV12'
	call	@div12')

define(`DIV12u',
``;;DIV12u'
	call	@div12u')

define(`ENTER',
``;;ENTER'
	call	@enter')

define(`EQ10f',
``;;EQ10f'
	mov	w,AL
	or	w,AH
	jz	@$+4
	jmp	@$1')

define(`EQ12',
``;;EQ12'
	call	@eq12')

define(`GE10f',
``;;GE10f'
	jnb	AH.7,@$+4
	jmp	@$1')

define(`GE12',
``;;GE12'
	call	@ge12')

define(`GE12u',
``;;GE12u'
	call	@ge12u')

define(`GETb1m',
``;;GETb1m($1)'
	mov	CL,#$1&255
	call	@getb1m')

define(`GETb1mu',
``;;GETb1mu($1)'
	mov	CL,#$1&255
	call	@getb1mu')

define(`GETb1p',
``;;GETb1p($1)'
	mov	CL,#$1&255
	call	@getb1p')

define(`GETb1pu',
``;;GETb1pu($1)'
	mov	CL,#$1&255
	call	@getb1pu')

define(`GETb1s',
``;;GETb1s($1)'
	mov	CL,#$1&255
	call	@getb1s')

define(`GETb1su',
``;;GETb1su($1)'
	mov	CL,#$1&255
	call	@getb1su')

define(`GETw1m',
``;;GETw1m($1)'
	mov	CL,#$1&255
	call	@getw1m')

define(`GETw1m_',
`GETw1m_PLUSn($1')

define(`GETw1m_PLUSn',
``;;GETw1m_PLUSn($1)'
	mov	CL,#($1)&255
`ifdef' _W_'$1`
	call	@getw1m
`else'
	call	@getb1m
`endif'
')

define(`GETw1n',
``;;GETw1n($1)'
ifelse($1, 0,
`	clr	AL
	clr	AH',
`	mov	AL,#$1&255
	mov	AH,#$1>>8')')

define(`GETw1p',
``;;GETw1p($1)'
	mov	CL,#$1&255
	call	@getw1p')

define(`GETw1s',
``;;GETw1s($1)'
	mov	CL,#$1&255
	call	@getw1s')

define(`GETw2m',
``;;GETw2m($1)'
	mov	CL,#$1&255
	call	@getw2m')

define(`GETw2n',
``;;GETw2n($1)'
ifelse($1, 0,
`	clr	BL
	clr	BH',
`	mov	BL,#$1&255
	mov	BH,#$1>>8')')

define(`GETw2p',
``;;GETw2p($1)'
	mov	CL,#$1&255
	call	@getw2p')

define(`GETw2s',
``;;GETw2s($1)'
	mov	CL,#$1&255
	call	@getw2s')

define(`GT10f',
``;;GT10f'
	jb	AH.7,@$+10
	mov	w,AL
	or	w,AH
	jz	@$+4
	jmp	@$1')

define(`GT12',
``;;GT12'
	call	@gt12')

define(`GT12u',
``;;GT12u'
	call	@gt12u')

define(`INCbp',
``;;INCbp'
	call	@incbp')

define(`INCwp',
``;;INCwp'
	call	@incwp')

define(`WORD_', 
`ifelse(LABELON, 1, `_WORD_)', `	dw	')')

define(`WORDn', 
`ifelse(LABELON, 1, `_WORDn, $1)', `	dw	$1')')

define(`WORDr0', 
`ifelse(LABELON, 1, `_WORDr0, $1)', `	dfw	$1')')

define(`JMPm',
``;;JMPm($1)'
	jmp	@$1')

define(`LABm',
`$1')

define(`LE10f',
``;;LE10f'
	jb	AH.7,@$+10
	mov	w,AL
	or	w,AH
	jz	@$+4
	jmp	@$1')

define(`LE12',
``;;LE12'
	call	@le12')

define(`LE12u',
``;;LE12u'
	call	@le12u')

define(`LNEG1',
``;;LNEG1'
	call	@lneg1')

define(`LT10f',
``;;LT10f'
	call	@lt10
	jnz	@$+4
	jmp	@$1')

define(`LT12',
``;;LT12'
	call	@lt12')

define(`LT12u',
``;;LT12u'
	call	@lt12u')

define(`MOD12',
``;;MOD12'
	call	@mod12')

define(`MOD12u',
``;;MOD12u'
	call	@mod12u')

define(`MOVE21',
``;;MOVE21'
	call	@move21')

define(`MUL12',
``;;NUL12'
	call	@mul12')

define(`MUL12u',
``;;NUL12u'
	call	@mul12u')

define(`NE10f',
``;;NE10f'
	mov	w,AL
	or	w,AH
	jnz	@$+4
	jmp	@$1')

define(`NE12',
``;;NE12'
	call	@ne12')

define(`NEARm',
`	dw	$1')

define(`OR12',
``;;OR12'
	call	@or12')

define(`PLUSn',
`+$1')

define(`POINT1l',
``;;POINT1l($1)'
	mov	AL,#($1)&255
	mov	AH,#($1)>>8')

define(`POINT1m',
``;;POINT1m($1)'
	mov	AL,#$1&255
	mov	AH,#$1>>8')

define(`POINT1s',
``;;POINT1s($1)'
	mov	AL,#$1&255
	add	AL,BP
	clr	AH')

define(`POINT2m',
``;;POINT2m($1)'
	mov	BL,#$1&255
	mov	BH,#$1>>8')

define(`POINT2m_',
`POINT2m_PLUSn($1')

define(`POINT2m_PLUSn',
``;;POINT2m_PLUSn($1)'
	mov	BL,#($1)&255
	mov	BH,#($1)>>8')

define(`POINT2s',
``;;POINT2s($1)'
	mov	BL,#$1&255
	add	BL,BP
	clr	BH')

define(`POP2',
``;;POP2'
	call	@pop2')

define(`PUSH1',
``;;PUSH1'
	call	@push1')

define(`PUSH2',
``;;PUSH2'
	call	@push2')

define(`PUSHm',
``;;PUSHm($1)'
	mov	CL,#$1&255
	csll	@pushm')

define(`PUSHp',
``;;PUSHp($1)'
	mov	CL,#$1&255
	call	@pushp')

define(`PUSHs',
``;;PUSHs($1)'
	mov	CL,#$1&255
	call	@pushs')

define(`PUT_m_',
`PUT_m_COMMAn($1')dnl

define(`PUT_m_COMMAn',
``;;PUT_m_COMMAn($1, $2)'
	mov	CL,#$1&255
	mov	DL,#$2&255
`ifdef' _W_'$1`
	mov	DH,#$2>>8
	call	@putwmn
`else'
	call	@putbmn
`endif'
')

define(`PUTbm1',
``;;PUTbm1($1)'
	mov	CL,#$1&255
	call	@putbm1')

define(`PUTbp1',
``;;PUTbp1'
	call	@putbp1')

define(`PUTwm1',
``;;PUTwm1'
	mov	CL,#$1&255
	call	@putwm1')

define(`PUTwp1',
``;;PUTwp1'
	call	@putwp1')

define(`rDEC1',
``;;rDEC1($1)'
	mov	DL,#$1&255
	mov	DH,#$1>>8
	call	@sub1n')

define(`rDEC2',
``;;rDEC2($1)'
	mov	DL,#$1&255
	mov	DH,#$1>>8
	call	@sub2n')

define(`RETURN',
``;;RETURN($1)'
	jmp	@return')

define(`rINC1',
``;;rINC1($1)'
	mov	DL,#$1&255
	mov	DH,#$1>>8
	call	@add1n')

define(`rINC2',
``;;rINC2($1)'
	mov	DL,#$1&255
	mov	DH,#$1>>8
	call	@add2n')

define(`SUB_m_',
`SUB_m_COMMAn($1')

define(`SUB_m_COMMAn',
``;;SUB_m_COMMAn($1, $2)'
	mov	CL,#$1&255
	mov	DL,#$2&255
`ifdef' _W_'$1`
	mov	DH,#$2>>8
	call	@subwmn
`else'
	call	@subbmn
`endif'
')

define(`SUB12',
``;;SUB12'
	call	@sub12')

define(`SUB1n',
``;;SUB1n($1)'
	mov	DL,#$1&255
	mov	DH,#$1>>8
	call	@sub1n')

define(`SUBbpn',
``;;SUBbpn'
	mov	DL,#$1&255
	call	@subbpn')

define(`SUBwpn',
``;;SUBwpn($1)'
	mov	DL,#$1&255
	mov	DH,#$1>>8
	call	@subwpn')

define(`SWAP12',
``;;SWAP12'
	call	@swap12')

define(`SWAP1s',
``;;SWAP1s'
	call	@swap1s')

define(`SWITCH',
``;;SWITCH'
	mov	CL,#($+6)&255
	mov	CH,#($+4)>>8
	call	@switch')

define(`XOR12',
``;;XOR12'
	call	@xor12')

define(`CSCINIT',
``;;CSCINIT'
	call	@_cscinit')

define(`ENTERM',
``;;ENTERM'
	;CLI')

define(`EXITM',
``;;EXITM'
	;STI')

divert
